#include "auxiliaryAxis.h"
#include "detectorTowerAxis.h"
#include "detectorTowerController.h"
#include "turboPmacController.h"
#include <epicsExport.h>
#include <errlog.h>
#include <iocsh.h>

/*
Contains all instances of auxiliaryAxis which have been created and is used
in the initialization hook function.
 */
static std::vector<auxiliaryAxis *> axes;

/**
 * @brief Hook function to perform certain actions during the IOC initialization
 *
 * @param iState
 */
static void epicsInithookFunction(initHookState iState) {
    if (iState == initHookAfterDatabaseRunning) {
        // Iterate through all axes of each and call the initialization method
        // on each one of them.
        for (std::vector<auxiliaryAxis *>::iterator itA = axes.begin();
             itA != axes.end(); ++itA) {
            auxiliaryAxis *axis = *itA;
            axis->init();
        }
    }
}

auxiliaryAxis::auxiliaryAxis(detectorTowerController *pC, int axisNo)
    : turboPmacAxis(pC, axisNo, false), pC_(pC) {

    /*
    The superclass constructor sinqAxis calls in turn its superclass constructor
    asynMotorAxis. In the latter, a pointer to the constructed object this is
    stored inside the array pAxes_:

        pC->pAxes_[axisNo] = this;

    Therefore, the axes are managed by the controller pC. If axisNo is out of
    bounds, asynMotorAxis prints an error (see
    https://github.com/epics-modules/motor/blob/master/motorApp/MotorSrc/asynMotorAxis.cpp,
    line 40). However, we want the IOC creation to stop completely, since this
    is a configuration error.
    */
    if (axisNo >= pC->numAxes()) {
        asynPrint(pC_->asynUserSelf(), ASYN_TRACE_ERROR,
                  "Controller \"%s\", axis %d => %s, line %d: FATAL ERROR: "
                  "Axis index %d must be smaller than the total number of axes "
                  "%d",
                  pC_->portName, axisNo_, __PRETTY_FUNCTION__, __LINE__,
                  axisNo_, pC->numAxes());
        exit(-1);
    }

    // Register the hook function during construction of the first axis object
    if (axes.empty()) {
        initHookRegister(&epicsInithookFunction);
    }

    // Collect all axes into this list which will be used in the hook
    // function
    axes.push_back(this);
}

auxiliaryAxis::~auxiliaryAxis(void) {
    // Since the controller memory is managed somewhere else, we don't need to
    // clean up the pointer pC here.
}

asynStatus auxiliaryAxis::init() {

    // Local variable declaration
    asynStatus status = asynSuccess;
    double motorRecResolution = 0.0;

    // The parameter library takes some time to be initialized. Therefore we
    // wait until the status is not asynParamUndefined anymore.
    time_t now = time(NULL);
    time_t maxInitTime = 60;
    while (1) {
        status = pC_->getDoubleParam(axisNo_, pC_->motorRecResolution(),
                                     &motorRecResolution);
        if (status == asynParamUndefined) {
            if (now + maxInitTime < time(NULL)) {
                asynPrint(pC_->asynUserSelf(), ASYN_TRACE_ERROR,
                          "Controller \"%s\", axis %d  => %s, line "
                          "%d\nInitializing the parameter library failed.\n",
                          pC_->portName, axisNo_, __PRETTY_FUNCTION__,
                          __LINE__);
                return asynError;
            }
        } else if (status == asynSuccess) {
            break;
        } else if (status != asynSuccess) {
            return pC_->paramLibAccessFailed(status, "motorRecResolution_",
                                             axisNo_, __PRETTY_FUNCTION__,
                                             __LINE__);
        }
    }
    return asynSuccess;
}

// Perform the actual poll
asynStatus auxiliaryAxis::doPoll(bool *moving) {

    // Return value for the poll
    asynStatus poll_status = asynSuccess;

    // Status of parameter library operations
    asynStatus pl_status = asynSuccess;

    int isMoving = 0;

    // =========================================================================

    /*
    The axis is moving if the detectorTowerAxis is moving -> We read the moving
    status from the detectorTowerAxis.
    */
    pl_status = pC_->getIntegerParam(dTA_->axisNo_, pC_->motorStatusMoving(),
                                     &isMoving);
    if (pl_status != asynSuccess) {
        return pC_->paramLibAccessFailed(pl_status, "motorStatusMoving_",
                                         dTA_->axisNo_, __PRETTY_FUNCTION__,
                                         __LINE__);
    }
    *moving = isMoving != 0;

    // Update the parameter library
    if (dTA_->error_ != 0) {
        pl_status = setIntegerParam(pC_->motorStatusProblem(), true);
        if (pl_status != asynSuccess) {
            return pC_->paramLibAccessFailed(pl_status, "motorStatusProblem_",
                                             axisNo_, __PRETTY_FUNCTION__,
                                             __LINE__);
        }
    }

    if (*moving == false) {
        pl_status = setIntegerParam(pC_->motorMoveToHome(), 0);
        if (pl_status != asynSuccess) {
            return pC_->paramLibAccessFailed(pl_status, "motorMoveToHome_",
                                             axisNo_, __PRETTY_FUNCTION__,
                                             __LINE__);
        }
    }

    pl_status = setIntegerParam(pC_->motorStatusMoving(), *moving);
    if (pl_status != asynSuccess) {
        return pC_->paramLibAccessFailed(pl_status, "motorStatusMoving_",
                                         axisNo_, __PRETTY_FUNCTION__,
                                         __LINE__);
    }

    pl_status = setIntegerParam(pC_->motorStatusDone(), !(*moving));
    if (pl_status != asynSuccess) {
        return pC_->paramLibAccessFailed(pl_status, "motorStatusDone_", axisNo_,
                                         __PRETTY_FUNCTION__, __LINE__);
    }

    // According to the function documentation of asynMotorAxis::poll, this
    // function should be called at the end of a poll implementation.
    pl_status = callParamCallbacks();
    bool wantToPrint = pl_status != asynSuccess;
    if (pC_->getMsgPrintControl().shouldBePrinted(
            pC_->portName, axisNo_, __PRETTY_FUNCTION__, __LINE__, wantToPrint,
            pC_->asynUserSelf())) {
        asynPrint(pC_->asynUserSelf(), ASYN_TRACE_ERROR,
                  "Controller \"%s\", axis %d => %s, line "
                  "%d:\ncallParamCallbacks failed with %s.%s\n",
                  pC_->portName, axisNo_, __PRETTY_FUNCTION__, __LINE__,
                  pC_->stringifyAsynStatus(poll_status),
                  pC_->getMsgPrintControl().getSuffix());
    }
    if (wantToPrint) {
        poll_status = pl_status;
    }

    // The limits are written into this class instance inside the doPoll
    // function of detectorTowerAxis
    return poll_status;
}

asynStatus auxiliaryAxis::doMove(double position, int relative,
                                 double min_velocity, double max_velocity,
                                 double acceleration) {

    double motorRecResolution = 0.0;
    asynStatus pl_status = pC_->getDoubleParam(
        axisNo_, pC_->motorRecResolution(), &motorRecResolution);
    if (pl_status != asynSuccess) {
        return pC_->paramLibAccessFailed(pl_status, "motorRecResolution_",
                                         axisNo_, __PRETTY_FUNCTION__,
                                         __LINE__);
    }

    // Signal to the deferredMovementCollectorLoop (of the detectorTowerAxis)
    // that a movement should be started to the defined target position.
    targetPosition_ = position * motorRecResolution;
    dTA_->receivedTarget_ = true;

    return asynSuccess;
}

asynStatus auxiliaryAxis::stop(double acceleration) {

    // Status of read-write-operations of ASCII commands to the controller
    asynStatus rw_status = asynSuccess;

    // Status of parameter library operations
    asynStatus pl_status = asynSuccess;

    char response[pC_->MAXBUF_];

    // =========================================================================

    rw_status = pC_->writeRead(axisNo_, "P350=8", response, 0);

    if (rw_status != asynSuccess) {
        asynPrint(
            pC_->asynUserSelf(), ASYN_TRACE_ERROR,
            "Controller \"%s\", axis %d  => %s, line %d\nStopping the movement "
            "failed.\n",
            pC_->portName, axisNo_, __PRETTY_FUNCTION__, __LINE__);

        pl_status = setIntegerParam(pC_->motorStatusProblem(), true);
        if (pl_status != asynSuccess) {
            return pC_->paramLibAccessFailed(pl_status, "motorStatusProblem_",
                                             axisNo_, __PRETTY_FUNCTION__,
                                             __LINE__);
        }
        return asynError;
    }

    return rw_status;
}

asynStatus auxiliaryAxis::reset() { return dTA_->reset(); };
