#include "detectorTowerAxis.h"
#include "detectorTowerController.h"
#include "turboPmacController.h"
#include <epicsExport.h>
#include <errlog.h>
#include <iocsh.h>
#include <math.h>

/*
Contains all instances of detectorTowerAxis which have been created and is used
in the initialization hook function.
 */
static std::vector<detectorTowerAxis *> axes;

/**
 * @brief Hook function to perform certain actions during the IOC initialization
 *
 * @param iState
 */
static void epicsInithookFunction(initHookState iState) {
    if (iState == initHookAfterDatabaseRunning) {
        // Iterate through all axes of each and call the initialization method
        // on each one of them.
        for (std::vector<detectorTowerAxis *>::iterator itA = axes.begin();
             itA != axes.end(); ++itA) {
            detectorTowerAxis *axis = *itA;
            axis->init();
        }
    }
}

static void deferredMovementCollectorLoop(void *drvPvt) {
    detectorTowerAxis *axis = (detectorTowerAxis *)drvPvt;
    while (1) {
        if (axis->receivedTarget_) {
            // Wait for 100 ms and then start the movement with the information
            // available
            axis->startingDeferredMovement_ = true;
            epicsThreadSleep(axis->deferredMovementWait_);
            axis->startCombinedMove();

            // After the movement command has been send, reset the flag
            axis->receivedTarget_ = false;
        }
        // Limit this loop to an idle frequency of 1 kHz
        epicsThreadSleep(0.001);
    }
}

detectorTowerAxis::detectorTowerAxis(detectorTowerController *pC, int axisNo,
                                     auxiliaryAxis *liftOffsetAxis,
                                     auxiliaryAxis *tiltOffsetAxis)
    : turboPmacAxis(pC, axisNo, false), pC_(pC) {

    /*
    The superclass constructor sinqAxis calls in turn its superclass constructor
    asynMotorAxis. In the latter, a pointer to the constructed object this is
    stored inside the array pAxes_:

        pC->pAxes_[axisNo] = this;

    Therefore, the axes are managed by the controller pC. If axisNo is out of
    bounds, asynMotorAxis prints an error (see
    https://github.com/epics-modules/motor/blob/master/motorApp/MotorSrc/asynMotorAxis.cpp,
    line 40). However, we want the IOC creation to stop completely, since this
    is a configuration error.
    */
    if (axisNo >= pC->numAxes()) {
        asynPrint(pC_->asynUserSelf(), ASYN_TRACE_ERROR,
                  "Controller \"%s\", axis %d => %s, line %d: FATAL ERROR: "
                  "Axis index %d must be smaller than the total number of axes "
                  "%d",
                  pC_->portName, axisNo_, __PRETTY_FUNCTION__, __LINE__,
                  axisNo_, pC->numAxes());
        exit(-1);
    }

    // Initialize all member variables
    // Assumed to be not ready by default, this is overwritten in the next poll
    error_ = 0;
    receivedTarget_ = false;
    startingDeferredMovement_ = false;
    deferredMovementWait_ = 0.1; // seconds
    liftOffsetAxis_ = liftOffsetAxis;
    tiltOffsetAxis_ = tiltOffsetAxis;
    liftOffsetAxis->dTA_ = this;
    tiltOffsetAxis->dTA_ = this;

    // Register the hook function during construction of the first axis object
    if (axes.empty()) {
        initHookRegister(&epicsInithookFunction);
    }

    // Collect all axes into this list which will be used in the hook
    // function
    axes.push_back(this);

    // Create a thread which collects all movement commands send to components
    // of the virtual axis. After a component received a new target position,
    // it forwards this information to the thread. The thread then waits a short
    // time to see if other components also received new targets and starts the
    // movement with all targets afterwards.
    epicsThreadCreate("deferredMovement", epicsThreadPriorityLow,
                      epicsThreadGetStackSize(epicsThreadStackMedium),
                      (EPICSTHREADFUNC)deferredMovementCollectorLoop,
                      (void *)this);
}

detectorTowerAxis::~detectorTowerAxis(void) {
    // Since the controller memory is managed somewhere else, we don't need to
    // clean up the pointer pC here.
}

asynStatus detectorTowerAxis::init() {

    // Local variable declaration
    asynStatus status = asynSuccess;
    double motorRecResolution = 0.0;

    // The parameter library takes some time to be initialized. Therefore we
    // wait until the status is not asynParamUndefined anymore.
    time_t now = time(NULL);
    time_t maxInitTime = 60;
    while (1) {
        status = pC_->getDoubleParam(axisNo_, pC_->motorRecResolution(),
                                     &motorRecResolution);
        if (status == asynParamUndefined) {
            if (now + maxInitTime < time(NULL)) {
                asynPrint(pC_->asynUserSelf(), ASYN_TRACE_ERROR,
                          "Controller \"%s\", axis "
                          "%ddeferredMovementCollectorLoop  => %s, line "
                          "%d\nInitializing the parameter library failed.\n",
                          pC_->portName, axisNo_, __PRETTY_FUNCTION__,
                          __LINE__);
                return asynError;
            }
        } else if (status == asynSuccess) {
            break;
        } else if (status != asynSuccess) {
            return pC_->paramLibAccessFailed(status, "motorRecResolution_",
                                             axisNo_, __PRETTY_FUNCTION__,
                                             __LINE__);
        }
    }

    return status;
}

// Perform the actual poll
asynStatus detectorTowerAxis::doPoll(bool *moving) {

    // Return value for the poll
    asynStatus poll_status = asynSuccess;

    // Status of read-write-operations of ASCII commands to the controller
    asynStatus rw_status = asynSuccess;

    // Status of parameter library operations
    asynStatus pl_status = asynSuccess;

    char response[pC_->MAXBUF_], userMessage[pC_->MAXBUF_];
    int nvals = 0;

    int direction = 0;
    int error = 0;
    int positionState = 0;
    int inPosition = 0;
    double beamTiltAngle = 0.0;
    double previousBeamTiltAngle = 0.0;
    double motorRecResolution = 0.0;
    double limitsOffset = 0.0;
    double highLimit = 0.0;
    double lowLimit = 0.0;
    double detectorHeight = 0.0;
    double detectorRadius = 0.0;
    double liftOffsetHighLimit = 0.0;
    double liftOffsetLowLimit = 0.0;
    double tiltOffsetHighLimit = 0.0;
    double tiltOffsetLowLimit = 0.0;

    // =========================================================================

    // Motor resolution from parameter library
    pl_status = pC_->getDoubleParam(axisNo_, pC_->motorRecResolution(),
                                    &motorRecResolution);
    if (pl_status != asynSuccess) {
        return pC_->paramLibAccessFailed(pl_status, "motorRecResolution_",
                                         axisNo_, __PRETTY_FUNCTION__,
                                         __LINE__);
    }

    // Read the previous motor position
    pl_status = motorPosition(&previousBeamTiltAngle);
    if (pl_status != asynSuccess) {
        return pl_status;
    }

    /*
    Query the following information:
    - In-position flag
    - Position state
    - Axis error
    - Beam tilt angle
    - Beam tilt angle limits
    - Beam lift offset limits
    - Detector tilt offset limits
    */
    const char *command = "P354 P358 P359 Q510 Q513 Q514 Q353 Q354 Q553 Q554";
    rw_status = pC_->writeRead(axisNo_, command, response, 10);
    if (rw_status != asynSuccess) {
        return rw_status;
    }

    nvals =
        sscanf(response, "%d %d %d %lf %lf %lf %lf %lf %lf %lf", &inPosition,
               &positionState, &error, &beamTiltAngle, &highLimit, &lowLimit,
               &liftOffsetHighLimit, &liftOffsetLowLimit, &tiltOffsetHighLimit,
               &tiltOffsetLowLimit);
    if (nvals != 10) {
        return pC_->couldNotParseResponse(command, response, axisNo_,
                                                __PRETTY_FUNCTION__, __LINE__);
    }

    /*
    The axis limits are set as: ({[]})
    where [] are the positive and negative limits set in EPICS/NICOS, {} are the
    software limits set on the MCU and () are the hardware limit switches. In
    other words, the EPICS/NICOS limits should be stricter than the software
    limits on the MCU which in turn should be stricter than the hardware limit
    switches. For example, if the hardware limit switches are at [-10, 10], the
    software limits could be at [-9, 9] and the EPICS / NICOS limits could be at
    [-8, 8]. Therefore, we cannot use the software limits read from the MCU
    directly, but need to shrink them a bit. In this case, we're shrinking them
    by limitsOffset on both sides.
    */
    pl_status =
        pC_->getDoubleParam(axisNo_, pC_->motorLimitsOffset(), &limitsOffset);
    if (pl_status != asynSuccess) {
        return pC_->paramLibAccessFailed(pl_status, "motorLimitsOffset_",
                                         axisNo_, __PRETTY_FUNCTION__,
                                         __LINE__);
    }
    highLimit = highLimit - limitsOffset;
    lowLimit = lowLimit + limitsOffset;

    // Store the axis status
    error_ = error;

    // Update the working position state PV
    pl_status = setIntegerParam(pC_->positionStateRBV(), positionState);
    if (pl_status != asynSuccess) {
        return pC_->paramLibAccessFailed(pl_status, "positionStateRBV_",
                                         axisNo_, __PRETTY_FUNCTION__,
                                         __LINE__);
    }

    // Axis is always enabled
    pl_status = setIntegerParam(pC_->motorEnableRBV(), 1);
    if (pl_status != asynSuccess) {
        return pC_->paramLibAccessFailed(pl_status, "motorEnableRBV_", axisNo_,
                                         __PRETTY_FUNCTION__, __LINE__);
    }

    // Check if the motor is moving
    if (inPosition == 1) {
        // By now, the controller has actually started the movement
        startingDeferredMovement_ = false;
        *moving = true;
    } else {
        if (startingDeferredMovement_) {
            *moving = true;
        } else {
            *moving = false;
        }
    }

    // Create the unique callsite identifier manually so it can be used later in
    // the shouldBePrinted calls.
    msgPrintControlKey keyPosState = msgPrintControlKey(
        pC_->portName, axisNo_, __PRETTY_FUNCTION__, __LINE__);

    // Reset the count, if the status is not an error state
    bool resetCountPosState = true;

    // Interpret the position state
    switch (positionState) {
    case 0:
        // Axis not ready
        break;
    case 1:
        // Axis ready
        break;
    case 2:
        // Axis is moving to or in changing position
        break;
    case 3:
        // Axis is moving to working state
        resetCountPosState = false;
        break;
    default:

        if (pC_->getMsgPrintControl().shouldBePrinted(keyPosState, true,
                                                      pC_->asynUserSelf())) {
            asynPrint(
                pC_->asynUserSelf(), ASYN_TRACE_ERROR,
                "Controller \"%s\", axis %d  => %s, line %d\nReached unknown "
                "state P354 = %d.%s\n",
                pC_->portName, axisNo_, __PRETTY_FUNCTION__, __LINE__,
                positionState, pC_->getMsgPrintControl().getSuffix());
        }
        resetCountPosState = false;

        snprintf(userMessage, sizeof(userMessage),
                 "Unknown state P354 = %d has been reached. Please call "
                 "the support.",
                 positionState);
        pl_status = setStringParam(pC_->motorMessageText(), userMessage);
        if (pl_status != asynSuccess) {
            return pC_->paramLibAccessFailed(pl_status, "motorMessageText_",
                                             axisNo_, __PRETTY_FUNCTION__,
                                             __LINE__);
        }
    }

    if (resetCountPosState) {
        pC_->getMsgPrintControl().resetCount(keyPosState, pC_->asynUserSelf());
    }

    if (*moving) {
        // If the axis is moving, evaluate the movement direction
        if ((beamTiltAngle - previousBeamTiltAngle) > 0) {
            direction = 1;
        } else {
            direction = 0;
        }
    }

    // Create the unique callsite identifier manually so it can be used later in
    // the shouldBePrinted calls.
    msgPrintControlKey keyError = msgPrintControlKey(
        pC_->portName, axisNo_, __PRETTY_FUNCTION__, __LINE__);
    bool resetError = true;

    // Error handling
    switch (error) {
    case 0:
        // No error
        break;
    case 1:
        if (pC_->getMsgPrintControl().shouldBePrinted(keyError, true,
                                                      pC_->asynUserSelf())) {
            asynPrint(
                pC_->asynUserSelf(), ASYN_TRACE_ERROR,
                "Controller \"%s\", axis %d  => %s, line %d\nBrake COZ not "
                "released (P359=1).%s\n",
                pC_->portName, axisNo_, __PRETTY_FUNCTION__, __LINE__,
                pC_->getMsgPrintControl().getSuffix());
        }
        resetError = false;

        pl_status =
            setStringParam(pC_->motorMessageText(),
                           "Brake COZ not released. Try resetting the axis. "
                           "If the problem persists, please call the support.");
        if (pl_status != asynSuccess) {
            return pC_->paramLibAccessFailed(pl_status, "motorMessageText_",
                                             axisNo_, __PRETTY_FUNCTION__,
                                             __LINE__);
        }

        poll_status = asynError;
        break;

    case 2:
        if (pC_->getMsgPrintControl().shouldBePrinted(keyError, true,
                                                      pC_->asynUserSelf())) {
            asynPrint(
                pC_->asynUserSelf(), ASYN_TRACE_ERROR,
                "Controller \"%s\", axis %d  => %s, line %d\nMove command "
                "rejected because axis is already moving.%s\n",
                pC_->portName, axisNo_, __PRETTY_FUNCTION__, __LINE__,
                pC_->getMsgPrintControl().getSuffix());
        }
        resetError = false;

        pl_status = setStringParam(pC_->motorMessageText(),
                                   "Move command rejected because axis is "
                                   "already moving. Please call the support");
        if (pl_status != asynSuccess) {
            return pC_->paramLibAccessFailed(pl_status, "motorMessageText_",
                                             axisNo_, __PRETTY_FUNCTION__,
                                             __LINE__);
        }

        poll_status = asynError;
        break;

    case 3:
        if (pC_->getMsgPrintControl().shouldBePrinted(keyError, true,
                                                      pC_->asynUserSelf())) {
            asynPrint(pC_->asynUserSelf(), ASYN_TRACE_ERROR,
                      "Controller \"%s\", axis %d  => %s, line %d\nError motor "
                      "FTZ.%s\n",
                      pC_->portName, axisNo_, __PRETTY_FUNCTION__, __LINE__,
                      pC_->getMsgPrintControl().getSuffix());
        }
        resetError = false;

        pl_status =
            setStringParam(pC_->motorMessageText(),
                           "Error in motor FTZ. Try resetting the axis. If "
                           "the problem persists, please call the support");
        if (pl_status != asynSuccess) {
            return pC_->paramLibAccessFailed(pl_status, "motorMessageText_",
                                             axisNo_, __PRETTY_FUNCTION__,
                                             __LINE__);
        }

        poll_status = asynError;
        break;

    case 4:
        if (pC_->getMsgPrintControl().shouldBePrinted(keyError, true,
                                                      pC_->asynUserSelf())) {
            asynPrint(
                pC_->asynUserSelf(), ASYN_TRACE_ERROR,
                "Controller \"%s\", axis %d  => %s, line %d\nAxis stopped "
                "unexpectedly.%s\n",
                pC_->portName, axisNo_, __PRETTY_FUNCTION__, __LINE__,
                pC_->getMsgPrintControl().getSuffix());
        }
        resetError = false;

        pl_status =
            setStringParam(pC_->motorMessageText(),
                           "Axis stopped unexpectedly. Try resetting the axis "
                           "and issue the move command again. If the problem "
                           "persists, please call the support.");
        if (pl_status != asynSuccess) {
            return pC_->paramLibAccessFailed(pl_status, "motorMessageText_",
                                             axisNo_, __PRETTY_FUNCTION__,
                                             __LINE__);
        }

        poll_status = asynError;
        break;

    case 5:
        if (pC_->getMsgPrintControl().shouldBePrinted(keyError, true,
                                                      pC_->asynUserSelf())) {
            asynPrint(
                pC_->asynUserSelf(), ASYN_TRACE_ERROR,
                "Controller \"%s\", axis %d  => %s, line %d\nRelease removed "
                "while moving.%s\n",
                pC_->portName, axisNo_, __PRETTY_FUNCTION__, __LINE__,
                pC_->getMsgPrintControl().getSuffix());
        }
        resetError = false;

        pl_status = setStringParam(
            pC_->motorMessageText(),
            "Axis release was removed while moving. Try resetting the axis "
            "and issue the move command again. If the problem persists, please "
            "call the support.");
        if (pl_status != asynSuccess) {
            return pC_->paramLibAccessFailed(pl_status, "motorMessageText_",
                                             axisNo_, __PRETTY_FUNCTION__,
                                             __LINE__);
        }

        poll_status = asynError;
        break;

    case 6:
        if (pC_->getMsgPrintControl().shouldBePrinted(keyError, true,
                                                      pC_->asynUserSelf())) {
            asynPrint(
                pC_->asynUserSelf(), ASYN_TRACE_ERROR,
                "Controller \"%s\", axis %d  => %s, line %d\nEmergency stop "
                "activated.%s\n",
                pC_->portName, axisNo_, __PRETTY_FUNCTION__, __LINE__,
                pC_->getMsgPrintControl().getSuffix());
        }
        resetError = false;

        pl_status =
            setStringParam(pC_->motorMessageText(), "Emergency stop activate");
        if (pl_status != asynSuccess) {
            return pC_->paramLibAccessFailed(pl_status, "motorMessageText_",
                                             axisNo_, __PRETTY_FUNCTION__,
                                             __LINE__);
        }

        poll_status = asynError;
        break;

    case 7:
        if (pC_->getMsgPrintControl().shouldBePrinted(keyError, true,
                                                      pC_->asynUserSelf())) {
            asynPrint(pC_->asynUserSelf(), ASYN_TRACE_ERROR,
                      "Controller \"%s\", axis %d  => %s, line %d\nError motor "
                      "COZ.%s\n",
                      pC_->portName, axisNo_, __PRETTY_FUNCTION__, __LINE__,
                      pC_->getMsgPrintControl().getSuffix());
        }
        resetError = false;

        pl_status =
            setStringParam(pC_->motorMessageText(),
                           "Error in motor COZ. Try resetting the axis. If "
                           "the problem persists, please call the support");
        if (pl_status != asynSuccess) {
            return pC_->paramLibAccessFailed(pl_status, "motorMessageText_",
                                             axisNo_, __PRETTY_FUNCTION__,
                                             __LINE__);
        }

        poll_status = asynError;
        break;

    case 8:
        if (pC_->getMsgPrintControl().shouldBePrinted(keyError, true,
                                                      pC_->asynUserSelf())) {
            asynPrint(pC_->asynUserSelf(), ASYN_TRACE_ERROR,
                      "Controller \"%s\", axis %d  => %s, line %d\nError motor "
                      "COM.%s\n",
                      pC_->portName, axisNo_, __PRETTY_FUNCTION__, __LINE__,
                      pC_->getMsgPrintControl().getSuffix());
        }
        resetError = false;

        pl_status =
            setStringParam(pC_->motorMessageText(),
                           "Error in motor COM. Try resetting the axis. If "
                           "the problem persists, please call the support");
        if (pl_status != asynSuccess) {
            return pC_->paramLibAccessFailed(pl_status, "motorMessageText_",
                                             axisNo_, __PRETTY_FUNCTION__,
                                             __LINE__);
        }

        poll_status = asynError;
        break;

    case 9:
        if (pC_->getMsgPrintControl().shouldBePrinted(keyError, true,
                                                      pC_->asynUserSelf())) {
            asynPrint(pC_->asynUserSelf(), ASYN_TRACE_ERROR,
                      "Controller \"%s\", axis %d  => %s, line %d\nError motor "
                      "COX.%s\n",
                      pC_->portName, axisNo_, __PRETTY_FUNCTION__, __LINE__,
                      pC_->getMsgPrintControl().getSuffix());
        }
        resetError = false;

        pl_status =
            setStringParam(pC_->motorMessageText(),
                           "Error in motor COX. Try resetting the axis. If "
                           "the problem persists, please call the support");
        if (pl_status != asynSuccess) {
            return pC_->paramLibAccessFailed(pl_status, "motorMessageText_",
                                             axisNo_, __PRETTY_FUNCTION__,
                                             __LINE__);
        }

        poll_status = asynError;
        break;

    case 10:
        if (pC_->getMsgPrintControl().shouldBePrinted(keyError, true,
                                                      pC_->asynUserSelf())) {
            asynPrint(pC_->asynUserSelf(), ASYN_TRACE_ERROR,
                      "Controller \"%s\", axis %d  => %s, line %d\nHit end "
                      "switch FTZ.%s\n",
                      pC_->portName, axisNo_, __PRETTY_FUNCTION__, __LINE__,
                      pC_->getMsgPrintControl().getSuffix());
        }
        resetError = false;

        pl_status =
            setStringParam(pC_->motorMessageText(), "Hit end switch FTZ");
        if (pl_status != asynSuccess) {
            return pC_->paramLibAccessFailed(pl_status, "motorMessageText_",
                                             axisNo_, __PRETTY_FUNCTION__,
                                             __LINE__);
        }

        poll_status = asynError;
        break;

    case 11:
        // Following error
        if (pC_->getMsgPrintControl().shouldBePrinted(keyError, true,
                                                      pC_->asynUserSelf())) {
            asynPrint(
                pC_->asynUserSelf(), ASYN_TRACE_ERROR,
                "Controller \"%s\", axis %d  => %s, line %d\nMaximum allowed "
                "following error FTZ exceeded.%s\n",
                pC_->portName, axisNo_, __PRETTY_FUNCTION__, __LINE__,
                pC_->getMsgPrintControl().getSuffix());
        }
        resetError = false;

        snprintf(userMessage, sizeof(userMessage),
                 "Maximum allowed following error exceeded (P359 = %d). "
                 "Check if movement range is blocked. "
                 "Otherwise please call the support.",
                 error);
        pl_status = setStringParam(pC_->motorMessageText(), userMessage);
        if (pl_status != asynSuccess) {
            return pC_->paramLibAccessFailed(pl_status, "motorMessageText_",
                                             axisNo_, __PRETTY_FUNCTION__,
                                             __LINE__);
        }

        poll_status = asynError;
        break;

    default:
        if (pC_->getMsgPrintControl().shouldBePrinted(keyError, true,
                                                      pC_->asynUserSelf())) {
            asynPrint(
                pC_->asynUserSelf(), ASYN_TRACE_ERROR,
                "Controller \"%s\", axis %d  => %s, line %d\nUnknown error "
                "P359 = %d.%s\n",
                pC_->portName, axisNo_, __PRETTY_FUNCTION__, __LINE__, error,
                pC_->getMsgPrintControl().getSuffix());
        }
        resetError = false;

        snprintf(userMessage, sizeof(userMessage),
                 "Unknown error P359 = %d. Please call the support.", error);
        pl_status = setStringParam(pC_->motorMessageText(), userMessage);
        if (pl_status != asynSuccess) {
            return pC_->paramLibAccessFailed(pl_status, "motorMessageText_",
                                             axisNo_, __PRETTY_FUNCTION__,
                                             __LINE__);
        }

        poll_status = asynError;
        break;
    }

    if (resetError) {
        pC_->getMsgPrintControl().resetCount(keyError, pC_->asynUserSelf());
    }

    // Update the parameter library
    if (error != 0) {
        pl_status = setIntegerParam(pC_->motorStatusProblem(), true);
        if (pl_status != asynSuccess) {
            return pC_->paramLibAccessFailed(pl_status, "motorStatusProblem_",
                                             axisNo_, __PRETTY_FUNCTION__,
                                             __LINE__);
        }
    }

    if (*moving == false) {
        pl_status = setIntegerParam(pC_->motorMoveToHome(), 0);
        if (pl_status != asynSuccess) {
            return pC_->paramLibAccessFailed(pl_status, "motorMoveToHome_",
                                             axisNo_, __PRETTY_FUNCTION__,
                                             __LINE__);
        }
    }

    pl_status = setIntegerParam(pC_->motorStatusMoving(), *moving);
    if (pl_status != asynSuccess) {
        return pC_->paramLibAccessFailed(pl_status, "motorStatusMoving_",
                                         axisNo_, __PRETTY_FUNCTION__,
                                         __LINE__);
    }

    pl_status = setIntegerParam(pC_->motorStatusDone(), !(*moving));
    if (pl_status != asynSuccess) {
        return pC_->paramLibAccessFailed(pl_status, "motorStatusDone_", axisNo_,
                                         __PRETTY_FUNCTION__, __LINE__);
    }

    pl_status = setIntegerParam(pC_->motorStatusDirection(), direction);
    if (pl_status != asynSuccess) {
        return pC_->paramLibAccessFailed(pl_status, "motorStatusDirection_",
                                         axisNo_, __PRETTY_FUNCTION__,
                                         __LINE__);
    }

    pl_status = pC_->setDoubleParam(axisNo_, pC_->motorHighLimitFromDriver(),
                                    highLimit);
    if (pl_status != asynSuccess) {
        return pC_->paramLibAccessFailed(pl_status, "motorHighLimitFromDriver_",
                                         axisNo_, __PRETTY_FUNCTION__,
                                         __LINE__);
    }

    pl_status =
        pC_->setDoubleParam(axisNo_, pC_->motorLowLimitFromDriver(), lowLimit);
    if (pl_status != asynSuccess) {
        return pC_->paramLibAccessFailed(pl_status, "motorLowLimitFromDriver_",
                                         axisNo_, __PRETTY_FUNCTION__,
                                         __LINE__);
    }

    // Set the values for the offset axes
    int liftAxisNo = liftOffsetAxis_->axisNo_;
    double liftOffsetPosition =
        detectorHeight + detectorRadius * sin(beamTiltAngle);

    pl_status = setMotorPosition(liftOffsetPosition);
    if (pl_status != asynSuccess) {
        return pl_status;
    }
    pl_status = pC_->setDoubleParam(liftAxisNo, pC_->motorHighLimitFromDriver(),
                                    liftOffsetHighLimit);
    if (pl_status != asynSuccess) {
        return pC_->paramLibAccessFailed(pl_status, "motorHighLimitFromDriver_",
                                         liftAxisNo, __PRETTY_FUNCTION__,
                                         __LINE__);
    }
    pl_status = pC_->setDoubleParam(liftAxisNo, pC_->motorLowLimitFromDriver(),
                                    liftOffsetLowLimit);
    if (pl_status != asynSuccess) {
        return pC_->paramLibAccessFailed(pl_status, "motorLowLimitFromDriver_",
                                         liftAxisNo, __PRETTY_FUNCTION__,
                                         __LINE__);
    }

    int tiltAxisNo = tiltOffsetAxis_->axisNo_;

    // There exists no readback value for tiltOffsetAxis_, hence we just set the
    // current position to the target position.
    tiltOffsetAxis_->setMotorPosition(tiltOffsetAxis_->targetPosition_);
    if (pl_status != asynSuccess) {
        return pl_status;
    }
    pl_status = pC_->setDoubleParam(tiltAxisNo, pC_->motorHighLimitFromDriver(),
                                    tiltOffsetHighLimit);
    if (pl_status != asynSuccess) {
        return pC_->paramLibAccessFailed(pl_status, "motorHighLimitFromDriver_",
                                         tiltAxisNo, __PRETTY_FUNCTION__,
                                         __LINE__);
    }
    pl_status = pC_->setDoubleParam(tiltAxisNo, pC_->motorLowLimitFromDriver(),
                                    tiltOffsetLowLimit);
    if (pl_status != asynSuccess) {
        return pC_->paramLibAccessFailed(pl_status, "motorLowLimitFromDriver_",
                                         tiltAxisNo, __PRETTY_FUNCTION__,
                                         __LINE__);
    }

    // Transform from motor to EPICS coordinates (see comment in
    // turboPmacAxis::init())
    pl_status = setMotorPosition(beamTiltAngle);
    if (pl_status != asynSuccess) {
        return pl_status;
    }

    pl_status = setIntegerParam(pC_->positionStateRBV(), positionState);
    if (pl_status != asynSuccess) {
        return pC_->paramLibAccessFailed(pl_status, "positionStateRBV_",
                                         axisNo_, __PRETTY_FUNCTION__,
                                         __LINE__);
    }

    return poll_status;
}

asynStatus detectorTowerAxis::doMove(double position, int relative,
                                     double min_velocity, double max_velocity,
                                     double acceleration) {
    double motorRecResolution = 0.0;
    asynStatus pl_status = pC_->getDoubleParam(
        axisNo_, pC_->motorRecResolution(), &motorRecResolution);
    if (pl_status != asynSuccess) {
        return pC_->paramLibAccessFailed(pl_status, "motorRecResolution_",
                                         axisNo_, __PRETTY_FUNCTION__,
                                         __LINE__);
    }

    // Signal to the deferredMovementCollectorLoop that a movement should be
    // started to the defined target position.
    targetPosition_ = position * motorRecResolution;
    receivedTarget_ = true;

    return asynSuccess;
}

asynStatus detectorTowerAxis::startCombinedMove() {

    // Status of read-write-operations of ASCII commands to the controller
    asynStatus rw_status = asynSuccess;

    // Status of parameter library operations
    asynStatus pl_status = asynSuccess;

    char command[pC_->MAXBUF_], response[pC_->MAXBUF_];
    double motorCoordinatesPosition = 0.0;
    int positionState = 0;

    // =========================================================================

    // If the axis is in changer position, it must be moved into working
    // position before any move can be started.
    bool isInChangerPos = positionState == 2 || positionState == 3;
    if (pC_->getMsgPrintControl().shouldBePrinted(
            pC_->portName, axisNo_, __PRETTY_FUNCTION__, __LINE__,
            isInChangerPos, pC_->asynUserSelf())) {
        asynPrint(pC_->asynUserSelf(), ASYN_TRACE_ERROR,
                  "Controller \"%s\", axis %d  => %s, line %d\nAxis cannot be "
                  "moved because it is moving from working to changer "
                  "position, is in changer position or is moving from changer "
                  "to working position.%s\n",
                  pC_->portName, axisNo_, __PRETTY_FUNCTION__, __LINE__,
                  pC_->getMsgPrintControl().getSuffix());
    }
    if (isInChangerPos) {
        startingDeferredMovement_ = false;
        return asynError;
    }

    // Set the target positions for beam tilt, detector tilt offset and lift
    // offset
    snprintf(command, sizeof(command), "Q451=%lf Q454=%lf Q456=%lf P350=1",
             targetPosition_, liftOffsetAxis_->targetPosition_,
             tiltOffsetAxis_->targetPosition_);

    // Lock the access to the controller since this function runs in another
    // thread than the poll method.
    pC_->lock();

    // We don't expect an answer
    rw_status = pC_->writeRead(axisNo_, command, response, 0);

    // Free the controller again
    pC_->unlock();

    if (rw_status != asynSuccess) {
        asynPrint(
            pC_->asynUserSelf(), ASYN_TRACE_ERROR,
            "Controller \"%s\", axis %d  => %s, line %d\nStarting movement to "
            "target position %lf failed.\n",
            pC_->portName, axisNo_, __PRETTY_FUNCTION__, __LINE__,
            motorCoordinatesPosition);
        pl_status = setIntegerParam(pC_->motorStatusProblem(), true);
        if (pl_status != asynSuccess) {
            return pC_->paramLibAccessFailed(pl_status, "motorStatusProblem_",
                                             axisNo_, __PRETTY_FUNCTION__,
                                             __LINE__);
        }
    }

    return rw_status;
}

asynStatus detectorTowerAxis::stop(double acceleration) {

    // Status of read-write-operations of ASCII commands to the controller
    asynStatus rw_status = asynSuccess;

    // Status of parameter library operations
    asynStatus pl_status = asynSuccess;

    char response[pC_->MAXBUF_];

    // =========================================================================

    rw_status = pC_->writeRead(axisNo_, "P350=8", response, 0);

    if (rw_status != asynSuccess) {
        asynPrint(
            pC_->asynUserSelf(), ASYN_TRACE_ERROR,
            "Controller \"%s\", axis %d  => %s, line %d\nStopping the movement "
            "failed.\n",
            pC_->portName, axisNo_, __PRETTY_FUNCTION__, __LINE__);

        pl_status = setIntegerParam(pC_->motorStatusProblem(), true);
        if (pl_status != asynSuccess) {
            return pC_->paramLibAccessFailed(pl_status, "motorStatusProblem_",
                                             axisNo_, __PRETTY_FUNCTION__,
                                             __LINE__);
        }
        return asynError;
    }

    return rw_status;
}

// The detector tower axis uses absolute encoders
asynStatus detectorTowerAxis::readEncoderType() {
    asynStatus status = setStringParam(pC_->encoderType(), AbsoluteEncoder);
    if (status != asynSuccess) {
        return pC_->paramLibAccessFailed(status, "encoderType_", axisNo_,
                                         __PRETTY_FUNCTION__, __LINE__);
    }
    return asynSuccess;
}

asynStatus
detectorTowerAxis::toggleWorkingChangerState(bool toChangingPosition) {

    char response[pC_->MAXBUF_];

    // Status of read-write-operations of ASCII commands to the controller
    asynStatus rw_status = asynSuccess;

    // Status of parameter library operations
    asynStatus pl_status = asynSuccess;

    bool moving = false;
    int positionState = 0;

    // =========================================================================

    doPoll(&moving);

    pl_status =
        pC_->getIntegerParam(axisNo_, pC_->positionStateRBV(), &positionState);
    if (pl_status != asynSuccess) {
        return pC_->paramLibAccessFailed(pl_status, "positionStateRBV_",
                                         axisNo_, __PRETTY_FUNCTION__,
                                         __LINE__);
    }

    if (moving) {
        asynPrint(pC_->asynUserSelf(), ASYN_TRACE_ERROR,
                  "Controller \"%s\", axis %d  => %s, line %d\nAxis is not "
                  "idle and can therefore not be moved to %s state.%s\n",
                  pC_->portName, axisNo_, __PRETTY_FUNCTION__, __LINE__,
                  toChangingPosition ? "changer" : "working",
                  pC_->getMsgPrintControl().getSuffix());

        pl_status = setStringParam(
            pC_->motorMessageText(),
            "Axis cannot be moved to changer position while it is moving.");
        if (pl_status != asynSuccess) {
            return pC_->paramLibAccessFailed(pl_status, "motorMessageText_",
                                             axisNo_, __PRETTY_FUNCTION__,
                                             __LINE__);
        }
        return asynError;
    }

    // Axis is already in the correct position
    bool isAlreadyThere = (toChangingPosition == false && positionState == 1) ||
                          (toChangingPosition == true && positionState == 2);

    if (isAlreadyThere) {
        asynPrint(pC_->asynUserSelf(), ASYN_TRACE_WARNING,
                  "Controller \"%s\", axis %d  => %s, line %d\nAxis is already "
                  "in %s position.%s\n",
                  pC_->portName, axisNo_, __PRETTY_FUNCTION__, __LINE__,
                  toChangingPosition ? "changer" : "working",
                  pC_->getMsgPrintControl().getSuffix());
        return asynSuccess;
    }

    // Move the axis into changer or working position
    if (toChangingPosition) {
        rw_status = pC_->writeRead(axisNo_, "P350=2", response, 0);
        pl_status = setStringParam(pC_->motorMessageText(),
                                   "Moving to changer position ...");
    } else {
        rw_status = pC_->writeRead(axisNo_, "P350=3", response, 0);
        pl_status = setStringParam(pC_->motorMessageText(),
                                   "Moving to working state ...");
    }
    if (pl_status != asynSuccess) {
        return pC_->paramLibAccessFailed(pl_status, "motorMessageText_",
                                         axisNo_, __PRETTY_FUNCTION__,
                                         __LINE__);
    }
    return rw_status;
}

asynStatus detectorTowerAxis::doReset() {

    char response[pC_->MAXBUF_];
    asynStatus pl_status = asynSuccess;
    int positionState = 0;

    pl_status =
        pC_->getIntegerParam(axisNo_, pC_->positionStateRBV(), &positionState);
    if (pl_status != asynSuccess) {
        return pC_->paramLibAccessFailed(pl_status, "positionStateRBV_",
                                         axisNo_, __PRETTY_FUNCTION__,
                                         __LINE__);
    }

    /*
    Check which action should be performed:
    - If positionState == 0 (not ready): P352 = 1 (Set axis in closed-loop
    mode)
    - If error_ == 10 or 11 (FTZ motor error): P352 = 3 (Recover FTZ)
    - If any other error: P352 = 2 (Reset error)
    */
    if (positionState == 0) {
        return pC_->writeRead(axisNo_, "P352=1", response, 0);
    } else if (error_ == 10 || error_ == 11) {
        return pC_->writeRead(axisNo_, "P352=3", response, 0);
    } else if (error_ != 0) {
        return pC_->writeRead(axisNo_, "P352=2", response, 0);
    } else {
        return asynSuccess;
    }
}

/*************************************************************************************/
/** The following functions are C-wrappers, and can be called directly from
 * iocsh */

extern "C" {

/*
C wrapper for the axis constructor. Please refer to the detectorTower
constructor documentation. The controller is read from the portName.
*/
asynStatus detectorTowerCreateAxis(const char *portName, int axisDetectorTower,
                                   int liftOffsetaxisNo, int tiltOffsetaxisNo) {

    /*
    findAsynPortDriver is a asyn library FFI function which uses the C ABI.
    Therefore it returns a void pointer instead of e.g. a pointer to a
    superclass of the controller such as asynPortDriver. Type-safe upcasting
    via dynamic_cast is therefore not possible directly. However, we do know
    that the void pointer is either a pointer to asynPortDriver (if a driver
    with the specified name exists) or a nullptr. Therefore, we first do a
    nullptr check, then a cast to asynPortDriver and lastly a (typesafe)
    dynamic_upcast to Controller
    https://stackoverflow.com/questions/70906749/is-there-a-safe-way-to-cast-void-to-class-pointer-in-c
     */
    void *ptr = findAsynPortDriver(portName);
    if (ptr == nullptr) {
        /*
        We can't use asynPrint here since this macro would require us
        to get a lowLevelPortUser_ from a pointer to an asynPortDriver.
        However, the given pointer is a nullptr and therefore doesn'taxis
        works w/o that, but doesn't offer the comfort provided
        by the asynTrace-facility
        */
        errlogPrintf("Controller \"%s\" => %s, line %d\nPort not found.",
                     portName, __PRETTY_FUNCTION__, __LINE__);
        return asynError;
    }
    // Unsafe cast of the pointer to an asynPortDriver
    asynPortDriver *apd = (asynPortDriver *)(ptr);

    // Safe downcast
    detectorTowerController *pC = dynamic_cast<detectorTowerController *>(apd);
    if (pC == nullptr) {
        errlogPrintf("Controller \"%s\" => %s, line %d\nController "
                     "is not a detectorTowerController.",
                     portName, __PRETTY_FUNCTION__, __LINE__);
        return asynError;
    }

    // Assert that the three indices are different from each other
    if (axisDetectorTower == liftOffsetaxisNo ||
        axisDetectorTower == tiltOffsetaxisNo ||
        tiltOffsetaxisNo == liftOffsetaxisNo) {
        errlogPrintf("Controller \"%s\" => %s, line %d\nAll three axis indices "
                     "must be unique.",
                     portName, __PRETTY_FUNCTION__, __LINE__);
        return asynError;
    }

    // Prevent manipulation of the controller from other threads while we
    // create the new axis.
    pC->lock();

    /*
    We create a new instance of the axis, using the "new" keyword to
    allocate it on the heap while avoiding RAII.
    https://github.com/epics-modules/motor/blob/master/motorApp/MotorSrc/asynMotorController.cpp
    https://github.com/epics-modules/asyn/blob/master/asyn/asynPortDriver/asynPortDriver.cpp

    The created object is registered in EPICS in its constructor and can safely
    be "leaked" here.
    */

    auxiliaryAxis *liftOffsetAxis = new auxiliaryAxis(pC, liftOffsetaxisNo);
    auxiliaryAxis *tiltOffsetAxis = new auxiliaryAxis(pC, tiltOffsetaxisNo);

#pragma GCC diagnostic ignored "-Wunused-but-set-variable"
#pragma GCC diagnostic ignored "-Wunused-variable"
    detectorTowerAxis *pAxis = new detectorTowerAxis(
        pC, axisDetectorTower, liftOffsetAxis, tiltOffsetAxis);

    // Allow manipulation of the controller again
    pC->unlock();
    return asynSuccess;
}

static const iocshArg CreateAxisArg0 = {"Controller name (e.g. mcu1)",
                                        iocshArgString};
static const iocshArg CreateAxisArg1 = {"Number of the detector angle axis",
                                        iocshArgInt};
static const iocshArg CreateAxisArg2 = {"Number of the lift offset axis",
                                        iocshArgInt};
static const iocshArg CreateAxisArg3 = {"Number of the tilt offset axis",
                                        iocshArgInt};
static const iocshArg *const CreateAxisArgs[] = {
    &CreateAxisArg0,
    &CreateAxisArg1,
    &CreateAxisArg2,
    &CreateAxisArg3,
};
static const iocshFuncDef configDetectorTowerCreateAxis = {"detectorTowerAxis",
                                                           4, CreateAxisArgs};
static void configDetectorTowerCreateAxisCallFunc(const iocshArgBuf *args) {
    detectorTowerCreateAxis(args[0].sval, args[1].ival, args[2].ival,
                            args[3].ival);
}

// =============================================================================

/**
 * @brief Set the setDeferredMovementWait (FFI implementation)
 *
 * @param portName                Name of the controller
 * @param axisNo                  Axis number
 * @param scaleMovTimeout         Scaling factor (in seconds)
 * @return asynStatus
 */
asynStatus setDeferredMovementWait(const char *portName, int axisNo,
                                   double deferredMovementWait) {

    sinqController *pC;
    pC = (sinqController *)findAsynPortDriver(portName);
    if (pC == nullptr) {
        errlogPrintf("Controller \"%s\" => %s, line %d:\nPort %s not found.",
                     portName, __PRETTY_FUNCTION__, __LINE__, portName);
        return asynError;
    }

    asynMotorAxis *asynAxis = pC->getAxis(axisNo);
    detectorTowerAxis *axis = dynamic_cast<detectorTowerAxis *>(asynAxis);
    if (axis == nullptr) {
        errlogPrintf("Controller \"%s\" => %s, line %d:\nAxis %d does not "
                     "exist or is not an instance of detectorTowerAxis.",
                     portName, __PRETTY_FUNCTION__, __LINE__, axisNo);
        return asynError;
    }

    axis->deferredMovementWait_ = deferredMovementWait;
    return asynSuccess;
}

static const iocshArg setDeferredMovementWaitArg0 = {"Controller port name",
                                                     iocshArgString};
static const iocshArg setDeferredMovementWaitArg1 = {"Axis number",
                                                     iocshArgInt};
static const iocshArg setDeferredMovementWaitArg2 = {
    "Minimum time a deferred movement will be delayed in order to collect "
    "commands in seconds",
    iocshArgDouble};
static const iocshArg *const setDeferredMovementWaitArgs[] = {
    &setDeferredMovementWaitArg0, &setDeferredMovementWaitArg1,
    &setDeferredMovementWaitArg2};
static const iocshFuncDef setDeferredMovementWaitDef = {
    "setDeferredMovementWait", 3, setDeferredMovementWaitArgs};

static void setDeferredMovementWaitCallFunc(const iocshArgBuf *args) {
    setDeferredMovementWait(args[0].sval, args[1].ival, args[2].dval);
}

// =============================================================================

// This function is made known to EPICS in detectorTower.dbd and is
// called by EPICS in order to register both functions in the IOC shell
static void detectorTowerAxisRegister(void) {
    iocshRegister(&configDetectorTowerCreateAxis,
                  configDetectorTowerCreateAxisCallFunc);
    iocshRegister(&setDeferredMovementWaitDef, setDeferredMovementWaitCallFunc);
}
epicsExportRegistrar(detectorTowerAxisRegister);

} // extern "C"
