#ifndef detectorTowerAxis_H
#define detectorTowerAxis_H
#include "auxiliaryAxis.h"
#include "turboPmacAxis.h"

// Forward declaration of the controller class to resolve the cyclic dependency
// between the controller and the axis .h-file. See
// https://en.cppreference.com/w/cpp/language/class.
class detectorTowerController;

class detectorTowerAxis : public turboPmacAxis {
  public:
    /**
     * @brief Construct a new detectorTowerAxis
     *
     * @param pController         Pointer to the associated controller
     * @param axisNo              Index of the axis
     * @param liftOffsetAxis      Pointer to the attached axis which controls
     * the lift offset
     * @param tiltOffsetAxis      Pointer to the attached axis which controls
     * the tilt offset
     */
    detectorTowerAxis(detectorTowerController *pController, int axisNo,
                      auxiliaryAxis *liftOffsetAxis,
                      auxiliaryAxis *tiltOffsetAxis);

    /**
     * @brief Destroy the detectorTowerAxis
     *
     */
    virtual ~detectorTowerAxis();

    /**
     * @brief Readout of some values from the controller at IOC startup
     *
     * @return asynStatus
     */
    asynStatus init();

    /**
     * @brief Implementation of the `stop` function from asynMotorAxis
     *
     * @param acceleration        Acceleration ACCEL from the motor record. This
     * value is currently not used.
     * @return asynStatus
     */
    asynStatus stop(double acceleration);

    /**
     * @brief Implementation of the `doPoll` function from sinqAxis. The
     * parameters are described in the documentation of `sinqAxis::doPoll`.
     *
     * @param moving
     * @return asynStatus
     */
    asynStatus doPoll(bool *moving);

    /**
     * @brief Set the target value for the detector angle and trigger a position
     * collection cycle
     *
     * @param position
     * @param relative
     * @param min_velocity
     * @param max_velocity
     * @param acceleration
     * @return asynStatus
     */
    asynStatus doMove(double position, int relative, double min_velocity,
                      double max_velocity, double acceleration);

    /**
     * @brief Start a movement to the target positions of this axis and the
     * attached offset axes.
     *
     * @return asynStatus
     */
    asynStatus startCombinedMove();

    /**
     * @brief This axis type has an absolute encoder by default
     *
     * @return asynStatus
     */
    asynStatus readEncoderType();

    /**
     * @brief If the input value is true, move the axis into working position,
     * otherwise into changer position.
     *
     * @return asynStatus
     */
    asynStatus toggleWorkingChangerState(bool toChangingPosition);

    /**
     * @brief Implementation of the `doReset` method of sinqAxis
     *
     * @param on
     * @return asynStatus
     */
    asynStatus doReset();

    // If true, either this axis or one of the auxiliaryAxis attached to it
    // received a movement command.
    bool receivedTarget_;

    // If set to true, the virtual axis is about to start a deferred movement
    // (but is currently still collecting movement commands from its component
    // axes)
    bool startingDeferredMovement_;

    /*
    The collector cycle waits until this time in seconds has passed before
    starting a deferred movement
    */
    double deferredMovementWait_;

  protected:
    detectorTowerController *pC_;
    int error_;
    auxiliaryAxis *tiltOffsetAxis_;
    auxiliaryAxis *liftOffsetAxis_;

  private:
    friend class auxiliaryAxis;
};

#endif
