#include "detectorTowerController.h"
#include "detectorTowerAxis.h"
#include "turboPmacController.h"
#include <epicsExport.h>
#include <errlog.h>
#include <iocsh.h>

/**
 * @brief Construct a new detectorTowerController object
 *
 * @param portName              See documentation of sinqController
 * @param ipPortConfigName      See documentation of sinqController
 * @param numAxes               See documentation of sinqController
 * @param movingPollPeriod      See documentation of sinqController
 * @param idlePollPeriod        See documentation of sinqController
 * @param comTimeout  Time after which a communication timeout error
 * is declared in writeRead (in seconds)
 * @param extraParams           See documentation of sinqController
 */
detectorTowerController::detectorTowerController(
    const char *portName, const char *ipPortConfigName, int numAxes,
    double movingPollPeriod, double idlePollPeriod, double comTimeout)
    : turboPmacController(portName, ipPortConfigName, numAxes, movingPollPeriod,
                          idlePollPeriod, NUM_detectorTower_DRIVER_PARAMS)

{

    asynStatus status = asynSuccess;

    status =
        createParam("POSITION_STATE_RBV", asynParamInt32, &positionStateRBV_);
    if (status != asynSuccess) {
        asynPrint(this->pasynUserSelf, ASYN_TRACE_ERROR,
                  "Controller \"%s\" => %s, line %d\nFATAL ERROR (creating a "
                  "parameter failed with %s).\nTerminating IOC",
                  portName, __PRETTY_FUNCTION__, __LINE__,
                  stringifyAsynStatus(status));
        exit(-1);
    }

    status = createParam("CHANGE_STATE", asynParamInt32, &changeState_);
    if (status != asynSuccess) {
        asynPrint(this->pasynUserSelf, ASYN_TRACE_ERROR,
                  "Controller \"%s\" => %s, line %d\nFATAL ERROR (creating a "
                  "parameter failed with %s).\nTerminating IOC",
                  portName, __PRETTY_FUNCTION__, __LINE__,
                  stringifyAsynStatus(status));
        exit(-1);
    }
}

asynStatus detectorTowerController::readInt32(asynUser *pasynUser,
                                              epicsInt32 *value) {

    // Check if the axis is a detectorTowerAxis
    detectorTowerAxis *axis = getDetectorTowerAxis(pasynUser);
    if (axis == nullptr) {
        // This is apparently a "normal" turboPmacAxis or an auxiliaryAxis
        return turboPmacController::readInt32(pasynUser, value);
    } else {
        // The detector tower cannot be disabled
        if (pasynUser->reason == motorCanDisable_) {
            *value = 0;
            return asynSuccess;
        } else {
            return turboPmacController::readInt32(pasynUser, value);
        }
    }
}

asynStatus detectorTowerController::writeInt32(asynUser *pasynUser,
                                               epicsInt32 value) {
    int function = pasynUser->reason;

    // =====================================================================

    detectorTowerAxis *axis = getDetectorTowerAxis(pasynUser);

    if (axis == nullptr) {
        // This is apparently a "normal" turboPmacAxis or an auxiliaryAxis
        return turboPmacController::writeInt32(pasynUser, value);
    } else {
        if (function == changeState_) {
            return axis->toggleWorkingChangerState(value);
        } else {
            return turboPmacController::writeInt32(pasynUser, value);
        }
    }
}

/*
Access one of the axes of the controller via the axis adress stored in asynUser.
If the axis does not exist or is not a Axis, a nullptr is returned and an
error is emitted.
*/
detectorTowerAxis *
detectorTowerController::getDetectorTowerAxis(asynUser *pasynUser) {
    asynMotorAxis *asynAxis = asynMotorController::getAxis(pasynUser);
    return dynamic_cast<detectorTowerAxis *>(asynAxis);
}

/*
Access one of the axes of the controller via the axis index.
If the axis does not exist or is not a Axis, the function must return Null
*/
detectorTowerAxis *detectorTowerController::getDetectorTowerAxis(int axisNo) {
    asynMotorAxis *asynAxis = asynMotorController::getAxis(axisNo);
    return dynamic_cast<detectorTowerAxis *>(asynAxis);
}

/*************************************************************************************/
/** The following functions are C-wrappers, and can be called directly from
 * iocsh */

extern "C" {

/*
C wrapper for the controller constructor. Please refer to the
detectorTowerController constructor documentation.
*/
asynStatus detectorTowerCreateController(const char *portName,
                                         const char *ipPortConfigName,
                                         int numAxes, double movingPollPeriod,
                                         double idlePollPeriod,
                                         double comTimeout) {
    /*
    We create a new instance of the controller, using the "new" keyword to
    allocate it on the heap while avoiding RAII.
    https://github.com/epics-modules/motor/blob/master/motorApp/MotorSrc/asynMotorController.cpp
    https://github.com/epics-modules/asyn/blob/master/asyn/asynPortDriver/asynPortDriver.cpp

    The created object is registered in EPICS in its constructor and can safely
    be "leaked" here.
    */
#pragma GCC diagnostic ignored "-Wunused-but-set-variable"
#pragma GCC diagnostic ignored "-Wunused-variable"
    detectorTowerController *pController = new detectorTowerController(
        portName, ipPortConfigName, numAxes, movingPollPeriod, idlePollPeriod,
        comTimeout);

    return asynSuccess;
}

/*
Define name and type of the arguments for the CreateController function
in the iocsh. This is done by creating structs with the argument names and
types and then providing "factory" functions
(configCreateControllerCallFunc). These factory functions are used to
register the constructors during compilation.
*/
static const iocshArg CreateControllerArg0 = {"Controller name (e.g. mcu1)",
                                              iocshArgString};
static const iocshArg CreateControllerArg1 = {"Asyn IP port name (e.g. pmcu1)",
                                              iocshArgString};
static const iocshArg CreateControllerArg2 = {"Number of axes", iocshArgInt};
static const iocshArg CreateControllerArg3 = {"Moving poll rate (s)",
                                              iocshArgDouble};
static const iocshArg CreateControllerArg4 = {"Idle poll rate (s)",
                                              iocshArgDouble};
static const iocshArg CreateControllerArg5 = {"Communication timeout (s)",
                                              iocshArgDouble};
static const iocshArg *const CreateControllerArgs[] = {
    &CreateControllerArg0, &CreateControllerArg1, &CreateControllerArg2,
    &CreateControllerArg3, &CreateControllerArg4, &CreateControllerArg5};
static const iocshFuncDef configDetectorTowerCreateController = {
    "detectorTowerController", 6, CreateControllerArgs};
static void
configDetectorTowerCreateControllerCallFunc(const iocshArgBuf *args) {
    detectorTowerCreateController(args[0].sval, args[1].sval, args[2].ival,
                                  args[3].dval, args[4].dval, args[5].dval);
}

// This function is made known to EPICS in turboPmac.dbd and is called by EPICS
// in order to register both functions in the IOC shell
static void detectorTowerControllerRegister(void) {
    iocshRegister(&configDetectorTowerCreateController,
                  configDetectorTowerCreateControllerCallFunc);
}
epicsExportRegistrar(detectorTowerControllerRegister);

} // extern "C"
